import typer
from recap.analyzers import create_analyzer
from recap.config import settings
from recap.paths import create_catalog_path
from recap.typing import AnalyzerInspector
from rich import print_json


app = typer.Typer()


@app.command()
def analyze(
    plugin: str = typer.Argument(...,
        help="Analyzer plugin name. Run `recap plugins analyzers` to get a full list.",
    ),
    url: str = typer.Argument(...,
        help="URL to analyze. URL configs in settings.toml will be used, if availabe.",
    ),
    path: str = typer.Argument(...,
        help="Path to analyze.",
    ),
):
    """
    Analyzes and returns metadata for a URL and path.
    """

    crawlers_configs = settings('crawlers', [])
    crawler_urls = set(map(
        lambda c: c.get('url'),
        crawlers_configs,
    ))

    if url and url not in crawler_urls:
        crawlers_configs.append({'url': url})

    for crawler_config in crawlers_configs:
        if url == crawler_config['url']:
            with create_analyzer(plugin, **crawler_config) as analyzer:
                inspector = AnalyzerInspector(type(analyzer))
                if catalog_path := create_catalog_path(
                    path,
                    *inspector.input_path_types(),
                ):
                    if metadata := analyzer.analyze(catalog_path):
                        metadata_dict = metadata.dict(
                            by_alias=True,
                            exclude_defaults=True,
                            exclude_none=True,
                            exclude_unset=True,
                        )
                        return_dict = metadata_dict.get(
                            '__root__',
                            metadata_dict,
                        )
                        print_json(data=return_dict, sort_keys=True)
                else:
                    raise ValueError(
                        f"Invalid path={path} for analyzer={plugin}",
                    )



from .abstract import AbstractCatalog
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path, PurePosixPath
from recap.config import RECAP_HOME, settings
from sqlalchemy import Column, DateTime, create_engine, Index, select, update
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.schema import Sequence
from sqlalchemy.sql import func, text
from sqlalchemy.types import BigInteger, Integer, String, JSON
from typing import Any, Generator
from urllib.parse import urlparse


DEFAULT_URL = f"sqlite:///{settings('root_path', RECAP_HOME)}/catalog/recap.db"
Base = declarative_base()


class CatalogEntry(Base):
    __tablename__ = 'catalog'

    # Sequence instead of autoincrement="auto" for DuckDB compatibility
    entry_id_seq = Sequence('entry_id_seq')
    id = Column(
        # Use Integer with SQLite since it's suggested by SQLalchemy
        BigInteger().with_variant(Integer, "sqlite"),
        entry_id_seq,
        primary_key=True,
    )
    parent = Column(String(65535), nullable=False)
    name = Column(String(4096), nullable=False)
    metadata_ = Column(
        'metadata',
        JSON().with_variant(JSONB, "postgresql"),
        nullable=False,
    )
    created_at = Column(
        DateTime,
        nullable=False,
        server_default=func.now(),
        index=True,
    )
    deleted_at = Column(DateTime)

    __table_args__ = (
        Index(
            'parent_name_idx',
            parent,
            name,
        ),
    )

    def is_deleted(self) -> bool:
        return self.deleted_at is not None


class DatabaseCatalog(AbstractCatalog):
    """
    DatabaseCatalog stores metadata entries in a `catalog` table using
    SQLAlchemy. The table has three main columns: parent, name, and metadata.
    The parent and name columns reflect the directory for the metadata (as
    defined by an AbstractBrowser). The metadata column contains a JSON blob of
    all the various metadata types and objects.

    Previous metadata versions are kept in `catalog` as well. A deleted_at
    field is used to tombstone deleted directories. Directories that were
    updated, not deleted, will not have a deleted_at set; there will just be a
    more recent row (as sorted by `id`).

    Reads return the most recent metadata that was written to the path. If the
    most recent record has a deleted_at tombstone, an None is returned.

    Search strings are simply passed along to the WHERE clause in a SELECT
    statement. This does leave room for SQL injection attacks; not thrilled
    about that.
    """

    def __init__(
        self,
        engine: Engine,
    ):
        self.engine = engine
        Base.metadata.create_all(engine)
        self.Session = sessionmaker(engine)

    def _clean_path(self, path: str) -> tuple[str, PurePosixPath]:
        path_posix = PurePosixPath('/', path)
        path_str = str(path_posix)
        return (path_str, path_posix)

    def touch(
        self,
        path: str,
    ):
        _, path_posix = self._clean_path(path)
        path_stack = list(path_posix.parts)
        cwd = '/'

        with self.Session() as session, session.begin():
            # Touch all parents to make sure they exist.
            while len(path_stack):
                cwd = PurePosixPath(cwd, *path_stack)

                # PurePosixPath('/').parts returns ('/',). We don't want to touch
                # the root because it doesn't fit the parent/name model that we
                # have.
                if len(cwd.parts) > 1:
                    maybe_row = session.scalar(
                        select(
                            CatalogEntry,
                        )
                        .filter(
                            CatalogEntry.parent == str(cwd.parent),
                            CatalogEntry.name == str(cwd.name),
                        ).order_by(
                            CatalogEntry.id.desc(),
                        )
                    )

                    if not maybe_row or maybe_row.is_deleted():
                        session.add(CatalogEntry(
                            parent=str(cwd.parent),
                            name=cwd.name,
                            metadata_={},
                        ))
                    else:
                        # Path exists and isn't deleted. We can assume all
                        # parents also exist, so no need to check.
                        break

                path_stack.pop()

    def write(
        self,
        path: str,
        metadata: dict[str, Any],
        patch: bool = True,
    ):
        path_str, path_posix = self._clean_path(path)
        self.touch(path_str)
        with self.Session() as session, session.begin():
            if patch:
                existing_doc = self._get_metadata(session, path_posix) or {}
                metadata = existing_doc | metadata
            session.add(CatalogEntry(
                parent=str(path_posix.parent),
                name=path_posix.name,
                metadata_=metadata,
            ))

    def rm(
        self,
        path: str,
    ):
        _, path_posix = self._clean_path(path)
        with self.Session() as session:
            session.execute(update(CatalogEntry).filter(
                # parent = /foo/bar/baz
                (CatalogEntry.parent == f"{path_posix}")
                # or parent = /foo/bar/baz/%
                | (CatalogEntry.parent.like(f"{path_posix}/%"))
                # or parent = /foo/bar and name = baz
                | (
                    (CatalogEntry.parent == str(path_posix.parent))
                    & (CatalogEntry.name == path_posix.name)
                )
            ).values(
                deleted_at = func.now()
            ).execution_options(synchronize_session=False))

    def ls(
        self,
        path: str,
        time: datetime | None = None,
    ) -> list[str] | None:
        path_str, _ = self._clean_path(path)
        with self.Session() as session:
            subquery = session.query(
                CatalogEntry.name,
                CatalogEntry.deleted_at,
                func.rank().over(
                    order_by=CatalogEntry.id.desc(),
                    partition_by=(
                        CatalogEntry.parent,
                        CatalogEntry.name,
                    )
                ).label('rnk')
            ).filter(
                CatalogEntry.parent == path_str,
                CatalogEntry.created_at <= (time or func.now()),
            ).subquery()
            query = session.query(subquery).filter(
                subquery.c.rnk == 1,
                subquery.c.deleted_at == None,
            )
            rows = session.execute(query).fetchall()
            return [row[0] for row in rows] or None

    def read(
        self,
        path: str,
        time: datetime | None = None,
    ) -> dict[str, Any] | None:
        _, path_posix = self._clean_path(path)
        with self.Session() as session:
            return self._get_metadata(session, path_posix, time)

    def search(
        self,
        query: str,
        time: datetime | None = None,
    ) -> list[dict[str, Any]]:
        with self.Session() as session:
            subquery = session.query(
                CatalogEntry.metadata_,
                CatalogEntry.deleted_at,
                func.rank().over(
                    order_by=CatalogEntry.id.desc(),
                    partition_by=(
                        CatalogEntry.parent,
                        CatalogEntry.name,
                    )
                ).label('rnk')
            ).filter(
                CatalogEntry.created_at <= (time or func.now()),
                # TODO Yikes. Pretty sure this is a SQL injection vulnerability.
                text(query)
            ).subquery()

            query = session.query(subquery).filter(
                subquery.c.rnk == 1,
                subquery.c.deleted_at == None,
            ) # pyright: ignore [reportGeneralTypeIssues]

            rows = session.execute(query).fetchall()

            return [row[0] for row in rows]

    def _get_metadata(
        self,
        session: Session,
        path_posix: PurePosixPath,
        time: datetime | None = None,
    ) -> Any | None:
        maybe_entry = session.scalar(
            select(
                CatalogEntry,
            ).where(
                CatalogEntry.parent == str(path_posix.parent),
                CatalogEntry.name == path_posix.name,
                CatalogEntry.created_at <= (time or func.now()),
            ).order_by(
                CatalogEntry.id.desc(),
            )
        )
        if maybe_entry and not maybe_entry.is_deleted():
            return maybe_entry.metadata_
        else:
            return None


@contextmanager
def create_catalog(
    url: str | None = None,
    engine: dict[str, Any] = {},
    **_,
) -> Generator['DatabaseCatalog', None, None]:
    if not url:
        # If no URL is set, default to SQLite
        url = DEFAULT_URL
        # Make sure the catalog directory exists
        db_path = urlparse(url).path # pyright: ignore [reportGeneralTypeIssues]
        Path(db_path).parent.mkdir(parents=True, exist_ok=True)
    yield DatabaseCatalog(create_engine(url, **engine))



import logging
import typer
from pathlib import PurePosixPath
from recap.browsers import create_browser
from recap.config import settings
from recap.plugins import load_browser_plugins
from rich import print_json


app = typer.Typer()
log = logging.getLogger(__name__)


@app.command()
def browse(
    url: str = typer.Argument(...,
        help="URL to browse. URL configs in settings.toml will be used, if availabe.",
    ),
    path: str = typer.Argument(
        '/',
        help="Path to list children for.",
    ),
):
    """
    Returns children for a URL and path.
    """

    clean_path = str(PurePosixPath(path))
    browser_plugins = load_browser_plugins()
    crawlers_configs = settings('crawlers', [])
    crawler_urls = set(map(
        lambda c: c.get('url'),
        crawlers_configs,
    ))

    if url and url not in crawler_urls:
        crawlers_configs.append({'url': url})

    for crawler_config in crawlers_configs:
        if url == crawler_config['url']:
            for browser_name in browser_plugins.keys():
                try:
                    with create_browser(browser_name, **crawler_config) as browser:
                        if children := browser.children(clean_path):
                            print_json(data=sorted([c.name() for c in children]))
                        else:
                            print_json(data=[])
                except Exception as e:
                        log.debug(
                            'Skipped browser for url=%s name=%s',
                            url,
                            browser_name,
                            exc_info=e,
                        )



from pathlib import PurePosixPath
from pydantic import BaseModel
from typing import ClassVar, Optional, Type
from starlette.routing import compile_path


class CatalogPath(BaseModel):
    """
    Represents a path in the data catalog.
    """

    template: ClassVar[str]
    """
    A list of templated paths that represent the various locations of this path
    type. The template format uses Starlette's simplified URI style. For
    example:

        /databases/{scheme}/instances/{instance}/schemas/{schema}
    """

    def name(self) -> str:
        return PurePosixPath(str(self)).name

    def __str__(self) -> str:
        """
        Find a template format that has parameters compatible with this
        instance. If one is found, fill it in with this instance's
        attributes and return it. For example, a template:

            /databases/{scheme}/instances/{instance}/schemas/{schema}

        Would return:

            /databases/postgresql/instances/localhost/schemas/public

        If the instance had:

            SomeCatalogPath(
                scheme='postgresql',
                instance='localhost',
                schema='public']
            )
        """
        # TODO Should use Starlette formatting so typed variables
        # (e.g. {path:path} or {name:str}) are handled properly.
        return self.template.format(**self.dict(
                by_alias=True,
                exclude_none=True,
                exclude_unset=True,
            ))

    @classmethod
    def from_path(cls, path: str) -> Optional['CatalogPath']:
        """
        Construct this model from a path if it matches the template.
        """
        path_str = str(path)
        template_regex, _, _ = compile_path(cls.template)
        if match := template_regex.match(path_str):
            return cls(**match.groupdict())
        return None


class RootPath(CatalogPath):
    """
    Canonical root path for the catalog.
    """
    template = '/'


def create_catalog_path(
    path: str,
    *types: Type[CatalogPath],
) -> CatalogPath | None:
    """
    Given a collection of CatalogPath types, find one that has a template that
    matches the provided path. If a matching path is found, instantiate the
    CatalogPath type and return it.
    """
    for type_ in types + (RootPath, ):
        if catalog_path := type_.from_path(path):
            return catalog_path
    return None



from datetime import datetime
from fastapi import APIRouter, Body, Depends, HTTPException
from pathlib import PurePosixPath
from pydantic import BaseModel
from recap.catalogs.abstract import AbstractCatalog
from recap.server import get_catalog
from typing import Any


router = APIRouter(
    prefix="/catalog",
)


@router.get(
    "/{path:path}/metadata",
    response_model_exclude_defaults=True,
    response_model_exclude_none=True,
    response_model_exclude_unset=True,
)
def read_metadata(
    path: str,
    time: datetime | None = None,
    catalog: AbstractCatalog = Depends(get_catalog),
) -> dict[str, Any]:
    print(clean_path(path))
    metadata = catalog.read(clean_path(path), time)
    if metadata:
        return metadata
    raise HTTPException(status_code=404)


@router.patch("/{path:path}/metadata")
def patch_metadata(
    path: str,
    metadata: dict[str, Any] = Body(),
    catalog: AbstractCatalog = Depends(get_catalog),
):
    catalog.write(clean_path(path), metadata, True)


@router.put("/{path:path}/metadata")
def put_metadata(
    path: str,
    metadata: dict[str, Any] = Body(),
    catalog: AbstractCatalog = Depends(get_catalog),
):
    catalog.write(clean_path(path), metadata)


@router.get("/{path:path}/children")
def list_children(
    path: str,
    time: datetime | None = None,
    catalog: AbstractCatalog = Depends(get_catalog),
) -> list[str]:
    children = catalog.ls(clean_path(path), time)
    if children:
        return children
    raise HTTPException(status_code=404)


@router.delete("/{path:path}")
def remove_directory(
    path: str,
    catalog: AbstractCatalog = Depends(get_catalog),
):
    catalog.rm(clean_path(path))


@router.get("")
def query_search(
    query: str,
    time: datetime | None = None,
    catalog: AbstractCatalog = Depends(get_catalog),
) -> list[dict[str, Any]]:
    return catalog.search(query, time)


def clean_path(path: str) -> str:
    return str(PurePosixPath('/', path))



import logging
import sqlalchemy
from .abstract import AbstractBrowser
from contextlib import contextmanager
from pydantic import Field
from recap.paths import CatalogPath, RootPath, create_catalog_path
from typing import Any, Callable, Generator, Union
from urllib.parse import urlparse


log = logging.getLogger(__name__)


class DatabaseRootPath(CatalogPath):
    scheme: str
    name_: str = Field(alias='name')
    template = '/databases/{scheme}/instances/{name}'


class SchemasPath(CatalogPath):
    template = '/schemas'


class SchemaPath(CatalogPath):
    schema_: str = Field(alias='schema')
    template = SchemasPath.template + '/{schema}'


class TablesPath(CatalogPath):
    schema_: str = Field(alias='schema')
    template = SchemaPath.template + '/tables'


class ViewsPath(CatalogPath):
    schema_: str = Field(alias='schema')
    template = SchemaPath.template + '/views'


class TablePath(CatalogPath):
    schema_: str = Field(alias='schema')
    table: str
    template = TablesPath.template + '/{table}'


class ViewPath(CatalogPath):
    schema_: str = Field(alias='schema')
    view: str
    template = ViewsPath.template + '/{view}'


DatabaseBrowserPath = Union[
    SchemasPath,
    SchemaPath,
    TablesPath,
    TablePath,
    ViewsPath,
    ViewPath,
]


class DatabaseBrowser(AbstractBrowser):
    """
    A browser that lists database objects. DatabaseBrowser uses SQLAlchemy and
    its supported dialects (https://docs.sqlalchemy.org/en/13/dialects/).

    DatabaseBrowser follows this directory sturcture:

        /databases/<scheme>/instances/<instance>/schemas/<schema>/tables/<table>
        /databases/<scheme>/instances/<instance>/schemas/<schema>/views/<view>

    The scheme is derived from the DB URL (e.g.
    postgres://user:pass@localhost/my_db would have scheme="postgres"). The
    instance defaults to the hostname, but can be overridden via the `name`
    config. Schema is whatever SQLAlchemy returns from `get_schema_names`.
    Ditto for table and view.

    This model does not make the distinction between schemas and databases as
    defined in standard information_schema formats
    (https://en.wikipedia.org/wiki/Information_schema). PostgreSQL, in
    particular, is a little weird because it has both (the schema is usually
    `public`).
    """

    def __init__(
        self,
        engine: sqlalchemy.engine.Engine,
        root_: DatabaseRootPath | None = None,
    ):
        self.engine = engine
        self.root_ = root_ or DatabaseBrowser.default_root(str(engine.url))

    def children(
        self,
        path: str,
    ) -> list[DatabaseBrowserPath] | None:
        catalog_path = create_catalog_path(
            path,
            *list(DatabaseBrowserPath.__args__), # pyright: ignore [reportGeneralTypeIssues]
        )
        match catalog_path:
            case RootPath():
                return [SchemasPath()]
            case SchemasPath():
                return [
                    SchemaPath(schema=s)
                    for s in self.schemas()
                ]
            case SchemaPath(schema_=schema):
                # TODO can we move this if to the case
                if schema in self.schemas():
                    return [
                        TablesPath(schema=schema),
                        ViewsPath(schema=schema),
                    ]
            case TablesPath(schema_=schema):
                return [
                    TablePath(schema=schema, table=t)
                    for t in self.tables(schema)
                ]
            case ViewsPath(schema_=schema):
                return [
                    ViewPath(schema=schema, view=v)
                    for v in self.views(schema)
                ]
        return None

    def schemas(self) -> list[str]:
        """
        :returns: All schema names in a database. In PostgreSQL, this is
            usually just `public`. In MySQL and others, it's usually a list of
            all database names.
        """

        return sqlalchemy.inspect(self.engine).get_schema_names()

    def tables(self, schema: str) -> list[str]:
        """
        :returns: All table names in a schema.
        """

        return self._tables_or_views(
            schema,
            sqlalchemy.inspect(self.engine).get_table_names,
        )

    def views(self, schema: str) -> list[str]:
        """
        :returns: All view names in a schema.
        """
        return self._tables_or_views(
            schema,
            sqlalchemy.inspect(self.engine).get_view_names,
        )

    def _tables_or_views(
        self,
        schema: str,
        get_method: Callable[[str], list[str]],
    ) -> list[str]:
        """
        Helper function that gets returns all tables or views for a given
        schema. This method exists because some DBs return `<schema>.<table>`
        and others just return `<table>`. To keep things standard, we strip out
        the `<schema>.` prefix if it exists.

        :param get_method: A SQLAlchemy inspection method; either
            (`get_table_names` or `get_schema_names`).
        :returns: All views or tables in a schema.
        """

        results = []
        try:
            for table_or_view in get_method(schema):
                # Stripe schema name from the table/view name. Some dialects
                # include the schema name as part of the table/view. Let's keep
                # things consistent.
                if table_or_view.startswith(f"{schema}."):
                    table_or_view = table_or_view[len(schema) + 1:]
                results.append(table_or_view)
        except Exception as e:
            # Just optimistically try, and ignore if we can't get info.
            # Easier than trying to figure out if permission exists.
            log.debug(
                'Unable to fetch tables or views for schema=%s',
                schema,
                exc_info=e,
            )
        return results

    def root(self) -> DatabaseRootPath:
        return self.root_

    @staticmethod
    def default_root(url: str) -> DatabaseRootPath:
        parsed_url = urlparse(url)
        # Given `posgrestql+psycopg2://foo:bar@baz/some_db`, return `postgresql`.
        scheme = parsed_url.scheme.split('+')[0]
        # Given `posgrestql+psycopg2://foo:bar@baz/some_db`, return `baz`.
        return DatabaseRootPath(
            scheme=scheme,
            name=parsed_url.netloc.split('@')[-1],
        )


@contextmanager
def create_browser(
    url: str,
    name: str | None = None,
    engine: dict[str, Any] = {},
    **_,
) -> Generator[DatabaseBrowser, None, None]:
    default_root = DatabaseBrowser.default_root(url)
    yield DatabaseBrowser(
        engine=sqlalchemy.create_engine(url, **engine),
        root_=DatabaseRootPath(
            scheme=default_root.scheme,
            name=name or default_root.name_,
        ),
    )



"""
This module dynamically adds all available browser paths to a router.

The module works using type hints. It inspects both the `children()` and
`analyze()` methods for each browser and analyzer. It then infers which
analyzers can analyze paths from each browser. The module uses this information
to construct Pydantic models that contain all compatible analyzers for each
path. It adds these paths and models to the router.
"""

import inspect
from datetime import datetime
from fastapi import APIRouter, Body, Depends, HTTPException
from pydantic import BaseModel
from recap.catalogs.abstract import AbstractCatalog
from recap.paths import CatalogPath
from recap.plugins import load_browser_plugins
from recap.server import get_catalog
from recap.typing import (
    BrowserInspector,
    ModuleInspector,
    get_pydantic_model_for_path,
)
from typing import Callable


router = APIRouter(
    prefix="/catalog",
)


def add_route(
    router: APIRouter,
    endpoint: Callable,
    method: str,
    browser_root_path: type[CatalogPath],
    child_path: type[CatalogPath],
    response_model: type[BaseModel] | None = None,
):
    """
    Adds a metadata endpoint to a router. The endpoint's signature is
    overwritten to include the fields from the root and child paths.

        root=/databases/{scheme}/instances/{name}
        child=/schemas/{schema}

    Would result in scheme, name, and schema parameters being added to the
    endpoint. This will cause FastAPI to pass these path parameters into the
    endpoint, which allows the *_metadata methods below to grab them in
    **kwargs.

    :param router: The router to add a route to.
    :param endpoint: The endpoint (method) to add to the router.
    :param method: GET, PUT, POST, PATCH, DELETE, and so on.
    :param browser_root_path: Browser's root path to be used in HTTP path.
    :param child_path: Child path relative to browser's root to be used in HTTP
        path.
    :param response_model: The response_model class for GET methods.
    """

    dynamic_params = []
    path_fields = browser_root_path.__fields__ | child_path.__fields__
    endpoint_signature = inspect.signature(endpoint)
    endpoint_existing_params = [
        p for p in endpoint_signature.parameters.values()
        if p.kind != inspect.Parameter.VAR_KEYWORD
    ]

    for field in path_fields.values():
        dynamic_params.append(
            inspect.Parameter(
                field.alias or field.name,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=field.type_,
                default=field.default,
            )
        )

    endpoint.__signature__ = inspect.Signature(
        dynamic_params + endpoint_existing_params
    )

    router.add_api_route(
        path=browser_root_path.template + child_path.template + '/metadata',
        endpoint=endpoint,
        dependencies=[Depends(get_catalog)],
        response_model=response_model,
        methods=[method],
        # These only apply for GET, but no harm in including them for all.
        response_model_exclude_defaults=True,
        response_model_exclude_none=True,
        response_model_exclude_unset=True,
    )


def add_routes(
    router: APIRouter,
    metadata_class: type[BaseModel],
    browser_root_path: type[CatalogPath],
    child_path: type[CatalogPath],
):
    """
    Helper method that adds GET, PUT, and PATCH routes to a router for a given
    model/browser path combination.

    :param router: The router to add routes to.
    :param metadata_class: The Pydantic metadata model that represents metadata
        for the root/child path.
    :param browser_root_path: Browser's root path to be used in HTTP path.
    :param child_path: Child path relative to browser's root to be used in HTTP
        path.
    """
    def read_metadata(
        time: datetime | None = None,
        catalog: AbstractCatalog = Depends(get_catalog),
        **kwargs,
    ) -> metadata_class:
        path = (
            browser_root_path.template
            + child_path.template
        ).format(**kwargs)
        metadata = catalog.read(path, time)
        if metadata:
            return metadata_class.parse_obj(metadata)
        raise HTTPException(status_code=404)

    def put_metadata(
        metadata: metadata_class = Body(),
        catalog: AbstractCatalog = Depends(get_catalog),
        **kwargs,
    ):
        path = (
            browser_root_path.template
            + child_path.template
        ).format(**kwargs)
        metadata_dict = metadata.dict(
            by_alias=True,
            exclude_defaults=True,
            exclude_none=True,
            exclude_unset=True,
        )
        catalog.write(path, metadata_dict)

    def patch_metadata(
        metadata: metadata_class = Body(),
        catalog: AbstractCatalog = Depends(get_catalog),
        **kwargs,
    ):
        path = (
            browser_root_path.template
            + child_path.template
        ).format(**kwargs)
        metadata_dict = metadata.dict(
            by_alias=True,
            exclude_defaults=True,
            exclude_none=True,
            exclude_unset=True,
        )
        catalog.write(path, metadata_dict, True)

    add_route(
        router,
        read_metadata,
        'GET',
        browser_root_path,
        child_path,
        metadata_class,
    )

    add_route(
        router,
        put_metadata,
        'PUT',
        browser_root_path,
        child_path,
    )

    add_route(
        router,
        patch_metadata,
        'PATCH',
        browser_root_path,
        child_path,
    )


def add_metadata_paths(router: APIRouter):
    """
    Dynamically adds all browser paths to the router. `add_metadata_paths`
    looks at the `children()` method's return CatalogPaths for each browser,
    and gets all analyzers that analyze one or more of those paths. The method
    then gets the return Pydantic models for each analyze and combines them
    into an uber-Metadata model. A GET/PUT/PATCH endpoint is added for each of
    these models.

    :param router: Router to add metadata routes to.
    """

    for browser_module in load_browser_plugins().values():
        module_inspector = ModuleInspector(browser_module)
        if browser_class := module_inspector.browser_type():
            browser_inspector = BrowserInspector(browser_class)
            root_path_type = browser_inspector.root_type()
            for child_path_class in browser_inspector.children_types():
                metadata_class = get_pydantic_model_for_path(child_path_class)
                if root_path_type and metadata_class:
                    add_routes(
                        router,
                        metadata_class,
                        root_path_type,
                        child_path_class,
                    )


add_metadata_paths(router)



import typer
from datetime import datetime
from recap import catalogs
from recap.config import settings
from rich import print_json


app = typer.Typer(help="Read and search the data catalog.")


@app.command()
def search(
    query: str,
    time: datetime = typer.Option(
        None, '--time', '-t',
        help=\
            "View metadata as of a point in time.",
    ),
):
    """
    Searches the data catalog. The query string syntax is dependent on the data
    catalog.
    """

    with catalogs.create_catalog(**settings('catalog', {})) as c:
        results = c.search(query, time)
        print_json(data=results, sort_keys=True)


@app.command("list")
def list_(
    path: str = typer.Argument('/'),
    time: datetime = typer.Option(
        None, '--time', '-t',
        help=\
            "View metadata as of a point in time.",
    ),
):
    """
    Lists a data catalog directory's children.
    """

    with catalogs.create_catalog(**settings('catalog', {})) as c:
        results = sorted(c.ls(path, time) or [])
        print_json(data=results)


@app.command()
def read(
    path: str,
    time: datetime = typer.Option(
        None, '--time', '-t',
        help=\
            "View metadata as of a point in time.",
    ),
):
    """
    Prints metadata from a path in the data catalog.
    """

    with catalogs.create_catalog(**settings('catalog', {})) as c:
        results = c.read(path, time) or []
        print_json(data=results, sort_keys=True)



from fastapi import FastAPI
from typing import Generator
from . import catalogs, plugins
from recap.catalogs.abstract import AbstractCatalog
from recap.config import settings


DEFAULT_URL = 'http://localhost:8000'


fastapp = FastAPI()


def get_catalog() -> Generator[AbstractCatalog, None, None]:
    with catalogs.create_catalog(**settings('catalog', {})) as c:
        yield c


@fastapp.on_event("startup")
def load_plugins():
    allowed_plugins = settings('server.plugins', [])
    router_plugins = plugins.load_router_plugins()
    for plugin_name, plugin_router in router_plugins.items():
        if not allowed_plugins or plugin_name in allowed_plugins:
            fastapp.include_router(plugin_router)



import logging
import sqlalchemy
from .columns import TableColumnAnalyzer, Columns
from contextlib import contextmanager
from pydantic import BaseModel
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.db import create_browser, TablePath, ViewPath
from typing import Generator


log = logging.getLogger(__name__)


class BaseColumnProfile(BaseModel):
    count: int


class BinaryColumnProfile(BaseColumnProfile):
    min_length: int | None
    max_length: int | None
    distinct: int | None
    nulls: int | None


class DateColumnProfile(BaseColumnProfile):
    min: str | None
    max: str | None
    distinct: int | None
    nulls: int | None
    unix_epochs: int | None


class NumericColumnProfile(BaseColumnProfile):
    min: int | float | None
    max: int | float | None
    average: int | float | None
    sum: int | float | None
    nulls: int | None
    zeros: int | None
    negatives: int | None


class StringColumnProfile(BaseColumnProfile):
    min_length: int | None
    max_length: int | None
    distinct: int | None
    nulls: int | None
    empty_strings: int | None


ColumnProfile = (
    BinaryColumnProfile |
    DateColumnProfile |
    NumericColumnProfile |
    StringColumnProfile |
    # This must be at the end, or the REST API might use it when encoding JSON
    # instead of more specific types.
    BaseColumnProfile
)


class Profile(BaseMetadataModel):
    __root__: dict[str, ColumnProfile] = {}


class TableProfileAnalyzer(AbstractAnalyzer):
    def __init__(self, engine: sqlalchemy.engine.Engine):
        self.engine = engine

    def analyze(
        self,
        path: TablePath | ViewPath,
    ) -> Profile | None:
        table = path.table if isinstance(path, TablePath) else path.view
        column_analyzer = TableColumnAnalyzer(self.engine)
        # TODO This is very proof-of-concept...
        # TODO ZOMG SQL injection attacks all over!
        # TODO Is db.Table().select the right way to paramaterize tables?
        columns = column_analyzer.analyze(path) or Columns()
        sql_col_queries = ''
        numeric_types = [
            'BIGINT', 'FLOAT', 'INT', 'INTEGER', 'NUMERIC', 'REAL', 'SMALLINT'
        ]
        date_types = [
            'DATE', 'DATETIME', 'TIMESTAMP'
        ]
        # TODO Excluding 'JSON' because PG's 'JSONB' doesn't have LENGTH()
        string_types = [
            'CHAR', 'CLOB', 'NCHAR', 'NVARCHAR', 'TEXT', 'VARCHAR'
        ]
        binary_types = [
            'BLOB', 'VARBINARY'
        ]
        stat_types = [
            'min',
            'max',
            'average',
            'sum',
            'distinct',
            'nulls',
            'zeros',
            'negatives',
            'min_length',
            'max_length',
            'empty_strings',
            'unix_epochs',
        ]

        with self.engine.connect() as conn:
            varchar_type = 'VARCHAR'

            # BigQuery doesn't havt FLOAT or VARCHAR, so use its type.
            # TODO SQLAlchemy should expose a dialect type for a generic type.
            if conn.dialect.name == 'bigquery':
                varchar_type = 'STRING'
            elif conn.dialect.name == 'snowflake':
                conn.execute("ALTER SESSION SET QUOTED_IDENTIFIERS_IGNORE_CASE = TRUE")

            for column_name, column in columns.dict()['__root__'].items():
                generic_type = column['generic_type']
                quoted_column_name = f'"{column_name}"'
                if conn.dialect.name in ['bigquery', 'mysql']:
                    quoted_column_name = f'`{column_name}`'
                if generic_type in numeric_types:
                    # TODO add approx median and quantiles
                    # TODO can we use a STRUCT or something here?
                    sql_col_queries += f"""
                        , MIN({quoted_column_name}) AS min_{column_name}
                        , MAX({quoted_column_name}) AS max_{column_name}
                        , AVG({quoted_column_name}) AS average_{column_name}
                        , SUM({quoted_column_name}) AS sum_{column_name}
                        , COUNT(DISTINCT {quoted_column_name}) AS {column_name}_distinct
                        , SUM(CASE WHEN {quoted_column_name} IS NULL THEN 1 ELSE 0 END) AS nulls_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} = 0 THEN 1 ELSE 0 END) AS zeros_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} < 0 THEN 1 ELSE 0 END) AS negatives_{column_name}
                    """
                if generic_type in string_types:
                    sql_col_queries += f"""
                        , MIN(LENGTH({quoted_column_name})) AS min_length_{column_name}
                        , MAX(LENGTH({quoted_column_name})) AS max_length_{column_name}
                        , COUNT(DISTINCT {quoted_column_name}) AS distinct_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} IS NULL THEN 1 ELSE 0 END) AS nulls_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} LIKE '' THEN 1 ELSE 0 END) AS empty_strings_{column_name}
                    """
                if generic_type in binary_types:
                    sql_col_queries += f"""
                        , MIN(LENGTH({quoted_column_name})) AS min_length_{column_name}
                        , MAX(LENGTH({quoted_column_name})) AS max_length_{column_name}
                        , COUNT(DISTINCT {quoted_column_name}) AS distinct_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} IS NULL THEN 1 ELSE 0 END) AS nulls_{column_name}
                    """
                if generic_type in date_types:
                    sql_col_queries += f"""
                        , CAST(MIN({quoted_column_name}) AS {varchar_type}) AS min_{column_name}
                        , CAST(MAX({quoted_column_name}) AS {varchar_type}) AS max_{column_name}
                        , COUNT(DISTINCT {quoted_column_name}) AS distinct_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} IS NULL THEN 1 ELSE 0 END) AS nulls_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} = TIMESTAMP '1970-01-01 00:00:00' THEN 1 ELSE 0 END) AS unix_epochs_{column_name}
                    """

            quoted_schema = f'"{path.schema_}"'
            quoted_table = f'"{table}"'
            if conn.dialect.name in ['bigquery', 'mysql']:
                quoted_schema = f'`{path.schema_}`'
                quoted_table = f'`{table}`'

            sql = f"""
                SELECT
                    COUNT(*) AS count {sql_col_queries}
                FROM
                    {quoted_schema}.{quoted_table} t
            """

            rows = conn.execute(sql)
            row = dict(rows.first() or {})
            results = {}

            for column_name in columns.dict()['__root__'].keys():
                col_stats = results.get(column_name, {'count': row['count']})
                for stat_type in stat_types:
                    stat_name = f"{stat_type}_{column_name}"
                    if stat_name in row:
                        stat_value = row[stat_name]
                        # JSON encoder can't handle decimal.Decimal
                        import decimal
                        if isinstance(row[stat_name], decimal.Decimal):
                            stat_value = float(row[stat_name])
                        col_stats[stat_type] = stat_value
                results[column_name] = col_stats

            results = {}
            for column_name, column in columns.__root__.items():
                generic_type = column.generic_type
                col_stats = {'count': row['count']}
                for stat_type in stat_types:
                    stat_name = f"{stat_type}_{column_name}"
                    if stat_name in row:
                        stat_value = row[stat_name]
                        # JSON encoder can't handle decimal.Decimal
                        import decimal
                        if isinstance(row[stat_name], decimal.Decimal):
                            stat_value = float(row[stat_name])
                        col_stats[stat_type] = stat_value
                if generic_type in numeric_types:
                    results[column_name] = NumericColumnProfile(**col_stats)
                elif generic_type in string_types:
                    results[column_name] = StringColumnProfile(**col_stats)
                elif generic_type in binary_types:
                    results[column_name] = BinaryColumnProfile(**col_stats)
                elif generic_type in date_types:
                    results[column_name] = DateColumnProfile(**col_stats)
                else:
                    results[column_name] = BaseColumnProfile(**col_stats)

            return Profile.parse_obj(results)


@contextmanager
def create_analyzer(
    **config,
) -> Generator['TableProfileAnalyzer', None, None]:
    with create_browser(**config) as browser:
        yield TableProfileAnalyzer(browser.engine)



from pydantic import BaseModel, create_model
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers import AbstractBrowser
from recap.paths import CatalogPath
from recap.plugins import load_analyzer_plugins
from types import ModuleType
from typing import Callable, get_type_hints, TypeVar


T = TypeVar('T')


class Inspector:
    def _method_types(
        self,
        param_name: str,
        method: Callable,
        base_type: type[T],
    ) -> list[type[T]]:
        """
        Given a method and a param name, look for all types in the param that
        extend base_type. Given this method:

            def foo() -> RootPath | DatabaseRootPath:
                ...
        
        Calling:

            _method_types('return', foo, CatalogPath)

        Would return [RootPath, DatabaseRootPath]. The method also handles list
        and dict args, so given:

            def bar(paths: list[SchemaPath | SchemasPath]):
                ...

        Calling:

            _method_types('paths', bar, CatalogPath)

        Would return [SchemaPath, SchemasPath].

        :param param_name: Param name as defined in get_type_hints response.
        :param method: Callable to inspect.
        :param base_type: Only return types that sublcass this type.
        """

        method_types = []
        type_hints = get_type_hints(method)
        if param_type_hints := type_hints.get(param_name):
            type_stack = [param_type_hints]
            while type_stack:
                child_type = type_stack.pop()
                try: # EAFP
                    # Not all types are classes, so issubclass will barf sometimes
                    if issubclass(child_type, base_type):
                        method_types.append(child_type)
                except:
                    pass
                if hasattr(child_type, '__args__'):
                    type_stack.extend(child_type.__args__) # pyright: ignore [reportGeneralTypeIssues]
        return method_types

    def _method_type(self, param_name: str, method: Callable, base_type: type[T]) -> type[T] | None:
        if types := self._method_types(
            param_name,
            method,
            base_type,
        ):
            return types[0]
        return None


class AnalyzerInspector(Inspector):
    def __init__(self, analyzer_class: type[AbstractAnalyzer]):
        self.analyzer_class = analyzer_class

    def input_path_types(self) -> list[type[CatalogPath]]:
        """
        Get analyzer.analyze()'s input path= parameter type(s).
        """

        return self._method_types(
            'path',
            self.analyzer_class.analyze,
            CatalogPath,
        )

    def return_type(self) -> type[BaseMetadataModel] | None:
        """
        Get analyzer.analyze()'s return type.
        """

        return self._method_type(
            'return',
            self.analyzer_class.analyze,
            BaseMetadataModel,
        )


class BrowserInspector(Inspector):
    def __init__(self, browser_class: type[AbstractBrowser]):
        self.browser_class = browser_class

    def children_types(self) -> list[type[CatalogPath]]:
        """
        Get browser.children()'s return type(s)
        """

        return self._method_types(
            'return',
            self.browser_class.children,
            CatalogPath,
        )

    def root_type(self) -> type[CatalogPath] | None:
        """
        Get browser.root()'s return type.
        """

        return self._method_type(
            'return',
            self.browser_class.root,
            CatalogPath,
        )


class ModuleInspector:
    def __init__(self, module: ModuleType):
        self.module = module

    def browser_type(self) -> type[AbstractBrowser] | None:
        """
        Get the return type for module.create_browser()
        """

        browser_module_return = get_type_hints(
            self.module.create_browser,
        ).get('return')
        if browser_module_return:
            # browser_module_return is Generator[SomeBrowserType, None, None].
            # __args__[0] gets SomeBrowserType from the Generator type.
            return browser_module_return.__args__[0]
        return None

    def analyzer_type(self) -> type[AbstractAnalyzer] | None:
        """
        Get the return type for module.create_analyzer()
        """

        analyzer_module_return = get_type_hints(
            self.module.create_analyzer,
        ).get('return')
        if analyzer_module_return:
            # analyzer_module_return is Generator[SomeAnalyzerType, None, None].
            # __args__[0] gets SomeAnalyzerType from the Generator type.
            return analyzer_module_return.__args__[0]
        return None


def get_analyzers_for_path(path_class: type[CatalogPath]) -> list[type[AbstractAnalyzer]]:
    analyzers = []
    for analyzer_module in load_analyzer_plugins().values():
        module_inspector = ModuleInspector(analyzer_module)
        if analyzer_class := module_inspector.analyzer_type():
            analyzer_inspector = AnalyzerInspector(analyzer_class)
            analyzer_input_paths = analyzer_inspector.input_path_types()
            if path_class in analyzer_input_paths:
                analyzers.append(analyzer_class)
    return analyzers


def get_pydantic_model_for_path(
    path_class: type[CatalogPath],
    name: str | None = None,
) -> type[BaseModel] | None:
    """
    Build a Pydantic model that combines all metadata models from analyzers
    compatible with `path_class`. Given:

        def analyze(path: SchemaPath) -> SomeSchemaMetadata:
            ...
        def analyze(path: SchemaPath) -> SomeOtherSchemaMetadata:
            ...

    Calling `get_pydantic_model_for_path(SchemaPath)` would return:

        class SchemaPathMetadata:
            some.schema: SomeSchemaMetadata
            some.other.schema: SomeOtherSchemaMetadata

    Where "some.schema" is SomeSchemaMetadata.key() and "some.other.schema" is
    SomeOtherSchemaMetadata.key().
    """

    path_attrs = {}
    for analyzer_class in get_analyzers_for_path(path_class):
        analyzer_inspector = AnalyzerInspector(analyzer_class)
        if analyzer_return_model := analyzer_inspector.return_type():
            path_attrs[analyzer_return_model.key()] = (analyzer_return_model, None)
    return create_model(
        name or f"{path_class.__name__}Metadata",
        **path_attrs,
    ) if path_attrs else None



from abc import ABC, abstractmethod
from pydantic import BaseModel
from recap.paths import CatalogPath


class BaseMetadataModel(BaseModel):
    @classmethod
    def key(cls: type['BaseMetadataModel']) -> str:
        _, parent, child = cls.__module__.rsplit('.', 2)
        return f"{parent}.{child}"


class AbstractAnalyzer(ABC):
    """
    The abstract class for all analyzers. Analyzers are responsible for
    inspecting data and returning metadata.
    """

    @abstractmethod
    def analyze(self, path: CatalogPath) -> BaseMetadataModel | None:
        """
        Analyze a path for an infrastructure instance. Only the path is
        specified because the URL for the instance is passed in via the config
        in `open()`.

        :returns: BaseMetadataModel Pydantic BaseModel, which represents
            discovered metadata. This data gets serialized as JSON in the
            catalog.
        """

        raise NotImplementedError



import logging.config
import tomli
from .config import settings
from pathlib import Path
from typing import Any


# Config key to define a custom TOML dictConfig location.
LOGGING_CONFIG_PATH = 'logging.config.path'

# Default logging config if no logging config path is set.
DEFAULT_LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'standard': {
            'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
        },
    },
    'handlers': {
        'default': {
            'formatter': 'standard',
            'class': 'rich.logging.RichHandler',
            'show_time': False,
            'show_level': False,
            'show_path': False,
        },
    },
    'loggers': {
        '': {
            'handlers': ['default'],
            'level': 'WARNING',
            'propagate': False
        },
        'recap': {
            'handlers': ['default'],
            'level': 'INFO',
            'propagate': False
        },
        'uvicorn': {
            'handlers': ['default'],
            'level': 'INFO',
            'propagate': False
        },
    }
}

def setup_logging() -> dict[str, Any]:
    """
    Configures logging. If a user defines a TOML file that conforms to Python's
    dictConfig schema:

    https://docs.python.org/3/library/logging.config.html#logging-config-dictschema

    Then the custom logging settings will be used. If no TOML file is
    configured, DEFAULT_LOGGING_CONFIG is used.
    """

    logging_config = DEFAULT_LOGGING_CONFIG
    logging_config_file_loc = settings(LOGGING_CONFIG_PATH)
    if logging_config_file_loc:
        logging_config_file_path = Path(logging_config_file_loc)
        logging_config_string = logging_config_file_path.read_text()
        logging_config = tomli.loads(logging_config_string)
    logging.config.dictConfig(logging_config)
    return logging_config



import logging
from contextlib import contextmanager
from pydantic import Field
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.db import create_browser, DatabaseRootPath, TablePath, ViewPath
from typing import Generator


log = logging.getLogger(__name__)


class Location(BaseMetadataModel):
    database: str
    instance: str
    # Schema is a reserved word in BaseModel.
    schema_: str = Field(alias='schema')
    # TODO Should validate either table or view is set and show in JSON schema.
    table: str | None = None
    view: str | None = None


class TableLocationAnalyzer(AbstractAnalyzer):
    def __init__(
        self,
        root: DatabaseRootPath,
    ):
        self.root = root

    def analyze(
        self,
        path: TablePath | ViewPath,
    ) -> Location | None:
        is_table = isinstance(path, TablePath)
        table = path.table if is_table else path.view
        table_or_view = 'table' if is_table else 'view'
        location_dict = {
            'database': self.root.scheme,
            'instance': self.root.name_,
            'schema': path.schema_,
            table_or_view: table,
        }
        return Location.parse_obj(location_dict)


@contextmanager
def create_analyzer(
    **config,
) -> Generator['TableLocationAnalyzer', None, None]:
    with create_browser(**config) as browser:
        yield TableLocationAnalyzer(browser.root())



import typer
from recap.config import settings
from recap.catalogs import create_catalog
from recap.crawler import create_crawler
from rich.progress import Progress, SpinnerColumn, TextColumn
from typing import Optional


app = typer.Typer()


@app.command()
def crawl(
    url: Optional[str] = typer.Argument(
        None,
        help=\
            "URL to crawl. If unset, all URLs in settings.toml are used.",
    ),
    analyzer_excludes: list[str] = typer.Option(
        [], '--exclude', '-e',
        help=\
            "Skip the specified analyzer when crawling.",
    ),
    path_filters: list[str] = typer.Option(
        [], '--filter', '-f',
        help=\
            "Crawl only certain paths. Format is Unix shell-style wildcards.",
    ),
):
    """
    Crawls infrastructure and writes metadata to the data catalog.
    """

    # Make sure URL is included in crawlers if it's passed in. This is needed
    # because Recap works with URLs even if no config was set for it in
    # settings.toml. It is smart enough to figure out the default module to
    # use.
    crawlers_configs = settings('crawlers', [])
    crawler_urls = set(map(
        lambda c: c.get('url'),
        crawlers_configs,
    ))

    if url and url not in crawler_urls:
        crawlers_configs.append({'url': url})

    if analyzer_excludes:
        for crawler_config in crawlers_configs:
            if not url or url == crawler_config['url']:
                crawler_config['excludes'] = analyzer_excludes

    if path_filters:
        for crawler_config in crawlers_configs:
            if not url or url == crawler_config['url']:
                crawler_config['filters'] = path_filters

    with create_catalog(**settings('catalog', {})) as catalog:
        for crawler_config in crawlers_configs:
            if not url or url == crawler_config['url']:
                with create_crawler(catalog, **crawler_config) as crawler:
                    spinner = SpinnerColumn(finished_text='[green]✓')
                    text = TextColumn("[progress.description]{task.description}")

                    with (Progress(spinner, text) as progress):
                        # Set up the spinner description.
                        task_id = progress.add_task(
                            description=f"Crawling {url} ...",
                            total=1,
                        )

                        crawler.crawl()

                        # Mark done, so we get a little green checkmark.
                        progress.update(task_id, completed=1)



import os
from dynaconf import Dynaconf
from pathlib import Path


RECAP_HOME = os.path.join(Path.home(), '.recap')
Path(RECAP_HOME).mkdir(parents=True, exist_ok=True)


settings = Dynaconf(
    envvar_prefix="RECAP",
    load_dotenv=True,
    root_path=RECAP_HOME,
    settings_files=['settings.toml', '.secrets.toml'],
)



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any


class AbstractCatalog(ABC):
    """
    The abstract class for all catalogs. Recap catalogs store metadata and
    expose read and search APIs. Catalogs follow the same directory structure
    as AbstractBrowsers.
    """

    @abstractmethod
    def touch(
        self,
        path: str,
    ):
        """
        Creates an empty directory.
        """

        raise NotImplementedError

    @abstractmethod
    def write(
        self,
        path: str,
        metadata: dict[str, Any],
        patch: bool = True,
    ):
        """
        Writes metadata to a directory location.

        :param type: Metadata types can be any string. Common examples are
            "schema", "pii", "indexes", "profile", and so on. Types are defined
            based on the data that AbstractAnalyzers return.
        """

        raise NotImplementedError

    @abstractmethod
    def rm(
        self,
        path: str,
    ):
        """
        Remove a directory or metadata entry. If type is note set, the whole
        directory (including all children and metadata) is removed.

        :param type: If specified, the type of metadata to delete.
        """

        raise NotImplementedError

    @abstractmethod
    def ls(
        self,
        path: str,
        time: datetime | None = None,
    ) -> list[str] | None:
        """
        Returns all children in a directory. This method does not signal
        whether or not a directory has metadata, since metadata is not a child
        of a directory. To check if a path has metadata, call `read`.

        :returns: A list of child names. Does not include absolute path.
        """

        raise NotImplementedError

    @abstractmethod
    def read(
        self,
        path: str,
        time: datetime | None = None,
    ) -> dict[str, Any] | None:
        """
        Read all metadata in a directory.

        :returns: Metadata dictionary of the format {"metadata_type": Any}.
        """

        raise NotImplementedError

    @abstractmethod
    def search(
        self,
        query: str,
        time: datetime | None = None,
    ) -> list[dict[str, Any]]:
        """
        Searches an entire catalog for metadata. The query syntax is dependent
        on the catalog implementation.

        :param query: A query string to match against metadata in a catalog.
        """

        raise NotImplementedError



from .abstract import AbstractAnalyzer
from contextlib import contextmanager
from recap.plugins import load_analyzer_plugins
from typing import Generator


@contextmanager
def create_analyzer(
    plugin: str,
    **config,
) -> Generator['AbstractAnalyzer', None, None]:
    analyzer_plugins = load_analyzer_plugins()
    if analyzer_module := analyzer_plugins.get(plugin):
        with analyzer_module.create_analyzer(**config) as analyzer:
            yield analyzer
